/**
  *  \file client/dialogs/vcrinfo.cc
  *  \brief VCR Info Dialog
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/vcrinfo.h"
#include "client/widgets/vcrinfohead.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets/button.h"
#include "client/widgets/vcrinfomain.h"
#include "client/widgets/helpwidget.h"
#include "game/classicvcr.h"
#include "client/vcrplay.h"
#include "ui/widgets/quit.h"
#include "game/shipquery.h"
#include "client/widgets/specview.h"

namespace {
    class WVcrInfo : public UIWindow {
        ObjectHolder h;
        UIButton* btn_goto;
        UIButton* btn_info;
        WVcrInfoHeader head;
        WVcrInfoMain* main;
        WVcrSelectorWindowGameInterface& gameif;
        void init();
        void onMove();
        void onGoTo();
        void doHullInfo();
     public:
        WVcrInfo(GClassicVcrDatabase& db, int slot, int side, WVcrSelectorWindowGameInterface& gameif);
        bool handleEvent(const UIEvent& event, bool second_pass);
    };
}

WVcrInfo::WVcrInfo(GClassicVcrDatabase& db, int slot, int side, WVcrSelectorWindowGameInterface& gameif)
    : UIWindow(_("Combat Information"), BLUE_WINDOW, COLOR_GRAY),
      btn_goto(0), btn_info(0),
      head(db.getBattle(slot), side),
      main(0),
      gameif(gameif)
{
    init();
}

void
WVcrInfo::init()
{
    // VBox
    //   WVcrInfoHeader
    //   HBox
    //     WVcrInfoMain
    //     VBox
    //       Spacer
    //       "S"
    //   HBox
    //     "Go to"
    //     "Back"
    //     "Tab"
    //     "Ins"
    //     "H"
    setLayout(UIVBoxLayout::instance5);

    main = &h.add(new WVcrInfoMain(false));
    add(head);

    btn_info = &h.add(new UIButton("S", 's'));
    UIGroup& g1 = h.add(new UIGroup(UIHBoxLayout::instance0));
    UIGroup& g11 = h.add(new UIGroup(UIVBoxLayout::instance5));
    g11.add(h.add(new UISpacer()));
    g11.add(*btn_info);
    g1.add(*main);
    g1.add(h.add(new UISpacer()));
    g1.add(g11);
    add(g1);

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    btn_goto = &h.add(new UIButton(_("Go to"), SDLK_RETURN));
    g.add(*btn_goto);
    g.add(h.add(new UIButton(0, _("Close"), SDLK_ESCAPE)).addStop());
    g.add(h.add(new UIButton(_("Tab - Other side"), SDLK_TAB)));
    g.add(h.add(new UIButton(_("Ins - Sim"), SDLK_INSERT)));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton("H", 'h')));
    add(g);
    add(h.add(new WHelpWidget("pcc2:vcrinfo")));
    add(h.add(new UIQuit(0)));

    head.getList().sig_moved.add(this, &WVcrInfo::onMove);
    pack();
    onMove();
}

void
WVcrInfo::onMove()
{
    WVcrHullList& list = head.getList();
    GVcrObject& obj = list.getVcrEntry().getObject(list.getSide(), false);
    main->set(list.getVcrEntry(), list.getSide(), list.getCurrentHull());
    btn_goto->setState(st_Disabled, !gameif.isGameObject(obj));
    btn_info->setState(st_Disabled, obj.isPlanet() || list.getCurrentHull() == 0);
}

void
WVcrInfo::onGoTo()
{
    const GVcrObject& obj = head.getList().getVcrEntry().getObject(head.getList().getSide(), false);
    if (gameif.isGameObject(obj)) {
        gameif.goToGameObject(obj);
        stop(1);
    }
}

void
WVcrInfo::doHullInfo()
{
    WVcrHullList& list = head.getList();
    GVcrObject& obj = list.getVcrEntry().getObject(list.getSide(), false);
    if (!obj.isPlanet() && list.getCurrentHull() != 0) {
        GShipQuery q;
        q.setHullId(list.getCurrentHull());
        q.setOwner(obj.getOwner());
        q.setPlayerDisplaySet(GPlayerSet(obj.getOwner()));
        q.setLevelDisplaySet(GExpLevelSet(obj.getExperienceLevel()));
        q.setCombatMass(obj.getMass(), 0);
        q.setCrew(obj.getCrew());
        WSpecView sv;
        sv.setQuery(q);
        sv.doStandardDialog();
    }
}

bool
WVcrInfo::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_TAB:
         case SDLK_TAB + ss_Shift:
            /* Swap sides */
            head.getList().setSide(!head.getList().getSide());
            return true;

         case SDLK_LEFT:
            head.getList().setSide(0);
            return true;

         case SDLK_RIGHT:
            head.getList().setSide(1);
            return true;

         case SDLK_INSERT:
            /* Add to sim, after fight */
            main->addToSim(true);
            return true;

         case SDLK_INSERT + ss_Shift:
            /* Add to sim, before fight */
            main->addToSim(false);
            return true;

         case SDLK_RETURN:
            onGoTo();
            return true;

         case 's':
            doHullInfo();
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

bool
doVcrInfo(GClassicVcrDatabase& db, int slot, int side, WVcrSelectorWindowGameInterface& gameif)
{
    WVcrInfo info(db, slot, side, gameif);
    ui_root->centerWidget(info);
    return ui_root->runModal(info) != 0;
}
