/**
  *  \file game/backup.cc
  *  \brief Backup File Creation
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/backup.h"
#include "cpluslib/utf8.h"
#include "cpluslib/format.h"
#include "arch/platform.h"
#include "io/dirs.h"
#include "util/except.h"
#include "game/pref.h"

ConfigStringOption opt_BackupTurn    (getUserPreferences(), "Backup.Turn");
ConfigStringOption opt_BackupResult  (getUserPreferences(), "Backup.Result");
ConfigStringOption opt_BackupUtil    (getUserPreferences(), "Backup.Util");
ConfigStringOption opt_BackupChart   (getUserPreferences(), "Backup.Chart");
ConfigStringOption opt_BackupScript  (getUserPreferences(), "Backup.Script");
ConfigStringOption opt_UnpackSource  (getUserPreferences(), "Unpack.Source");    // FIXME: not implemented!
ConfigStringOption opt_MaketurnTarget(getUserPreferences(), "Maketurn.Target");

/** Try to create a path. Creates a complete path that can contain multiple non-existant
    directory levels. This does not fail when the path cannot be created; in that case,
    subsequent operations using the path will fail.
    \param dirName Name of path to create */
static void
tryCreatePath(const string_t dirName)
{
    const string_t parentName = Directory::getDirName(dirName);
    const string_t childName  = Directory::getFileName(dirName);

    // If parentName is the same as dirName, this means that dirName does not have a parent.
    // In this case, we don't do anything.
    if (parentName != dirName) {
        // Try enumerating the parent's content. If that fails, try to create it.
        // (openDir alone does not check whether the directory actually exists.)
        try {
            Ptr<Directory> parent = Directory::openDir(parentName);
            parent->getEntries();
        }
        catch (FileProblemException&) {
            tryCreatePath(parentName);
        }

        // Parent should now exist. Try creating child in it unless it already exists.
        try {
            Ptr<Directory> parent = Directory::openDir(parentName);
            Ptr<DirectoryEntry> entry = parent->getEntryByName(childName);
            if (entry->getFileType() != DirectoryEntry::FT_DIRECTORY) {
                entry->createAsDirectory();
            }
        }
        catch (FileProblemException&) { }
    }
}


/************************** GBackupFileTemplate **************************/

/** Create blank template. */
GBackupFileTemplate::GBackupFileTemplate()
    : game_dir(),
      player_nr(1),
      turn_nr(1)
{ }

/** Expand a template. Returns the expanded directory name. */
string_t
GBackupFileTemplate::expandFileName(string_t tpl)
{
    string_t result;
    Utf8Reader rdr(tpl);
    while (rdr.hasMore()) {
        uint32_t ch = rdr.get();
        if (ch == '%' && rdr.hasMore()) {
            ch = rdr.get();
            switch (ch) {
             case 'd': case 'D':
                // Directory
                if (result.empty()) {
                    // This is the first thing
                    result = game_dir;
                } else {
                    // Not the first thing; append just the basename of game_dir
                    result += Directory::getFileName(Directory::getAbsoluteName(game_dir));
                }

                // Make sure that the name ends in a path separator
                if (!result.empty() && !Directory::isSeparatorChar(result[result.size()-1])) {
                    result += Platform::PATH_SEPARATOR;
                }

                // Make sure that the template does not produce another path separator
                if (rdr.hasMore() && Directory::isSeparatorChar(*rdr.getRemainingData())) {
                    rdr.get();
                }
                break;
             case 'p': case 'P':
                // Player number
                result += format("%d", player_nr);
                break;
             case 't': case 'T':
                // Turn number
                result += format("%03d", turn_nr);
                break;
             case '%':
             default:
                appendUtf8(result, ch);
            }
        } else {
            appendUtf8(result, ch);
        }
    }
    return result;
}

/** Copy a file, using a template.
    \param tpl Template file name (can be empty; in this case, the function does nothing)
    \param src Open file
    \param pm Progress monitor */
void
GBackupFileTemplate::copyFile(string_t tpl, Stream& src, ProgressMonitor& pm)
{
    // Anything to do?
    if (tpl.empty()) {
        return;
    }

    // Create directory for file
    string_t name = expandFileName(tpl);
    tryCreatePath(Directory::getDirName(name));

    // Do it
    Ptr<Stream> file = Stream::openFile(name, Stream::C_CREATE);
    file->copyFrom(src, pm);
}

/** Erase a file, using a template. */
void
GBackupFileTemplate::eraseFile(string_t tpl)
{
    // Anything to do?
    if (tpl.empty()) {
        return;
    }

    // Erase the file
    string_t name   = expandFileName(tpl);
    string_t parent = Directory::getDirName(name);
    string_t child  = Directory::getFileName(name);
    try {
        Ptr<Directory> dir = Directory::openDir(parent);
        dir->eraseNT(child);
    }
    catch (FileProblemException&) { }
}
