/*
 *  TEST: VcrDatabase, VCR players
 *
 *  This takes up to three parameters:
 *  - a VCR file
 *  - a spec file directory (defaults to /usr/games/planets)
 *  - a repeat count (defaults to 1)
 *  The VCR file is read and played; results are displayed on stdout.
 */
#include <iostream>
#include <cstdlib>
#include <vector>
#include <iomanip>
#include "game/struct.h"
#include "game/tvcr.h"
#include "io/streams.h"
#include "util/except.h"
#include "util/console.h"
#include "game/global.h"
#include "game/spec.h"
#include "game/config.h"
#include "game/vcrdb.h"
#include "game/player.h"
#include "game/racename.h"
#include "game/gamedirs.h"
#include "util/modinit.h"
#include "arch/platform.h"
#include "game/team.h"
#include "game/classicvcr.h"

static const char*const results[] = {
    "left-destroyed",
    "right-destroyed",
    "left-captured",
    "right-captured",
    "timeout"
};

void
help()
{
    std::cout << "Usage: playvcr vcrfile [rootdir [repeat]]\n";
    std::exit(1);
}

Ptr<GClassicVcrDatabase>
loadVcrs(const char* name)
{
    Ptr<Stream> s = Stream::openFile(name, Stream::C_READ);
    Ptr<GClassicVcrDatabase> db = new GClassicVcrDatabase(*s);
    std::cout << "VCR file contains " << db->getNumBattles() << " entries\n";
    return db;
}

static const ModuleInitializer* testvcr_modules[] = {
    mod_base_console,
    mod_team,
    mod_spec,
    mod_config,
    0
};

int
main(int, char** argv)
{
    performInitialisation(OnStartup, testvcr_modules);
    argv = Platform::convertArgumentsToUtf8(argv);

    char* file = 0;
    char* dir = 0;
    bool repeat_set = false;
    int repeat = 1;

    while (*++argv) {
        if (!file)
            file = *argv;
        else if (!dir)
            dir = *argv;
        else if (!repeat_set) {
            repeat_set = true;
            repeat = std::atoi(*argv);
            if (repeat <= 0)
                help();
        } else
            help();
    }

    if (!file)
        help();

    try {    
        initDirectories(dir ? dir : ".", "/usr/games/planets");
        performInitialisation(OnEnterDirectory | OnEnterGame, testvcr_modules);
        
        Ptr<GClassicVcrDatabase> vcrs = loadVcrs(file);
        
        std::cout << "--- Starting Playback ---" << std::endl;

        for (int rc = 0; rc < repeat; ++rc) {
            for(int i = 0; i < vcrs->getNumBattles(); ++i) {
                GClassicVcrEntry& en = vcrs->getBattle(i);
                std::cout << "Record #" << (i+1) << ":\n";

                Ptr<VcrPlayer> ply = en.createPlayer();
                ply->setVisualizer(new VcrNullVisualizer());
                
                GClassicVcr the_vcr = en.getVcr();
                if (!ply->setCapabilities(en.getCapabilities())) {
                    std::cout << "\tinvalid (cap)\n";
                    continue;
                }
                if (ply->checkVcr(the_vcr)) {
                    std::cout << "\tinvalid (checkVcr)\n";
                    continue;
                }
                ProgressMonitor pm;
                ply->playVcr(the_vcr, pm);
                
                GClassicVcr result;
                ply->doneVcr(result);
                int status = ply->getStatusWord();
                int time = ply->getTime();
                std::cout << "\tEnding time " << time
                          << " (" << time/60 << ":" << std::setw(2) << std::setfill('0') << time%60
                          << ")\n";
                if (status == VcrPlayer::VCRS_INVALID) {
                    std::cout << "\tinvalid (play)\n";
                } else {
                    char t = '\t';
                    for (int i = 0; i < 5; ++i) {
                        if (status & (1 << i)) {
                            std::cout << t << results[i];
                            t = ' ';
                        }
                    }
                    if (t == '\t')
                        std::cout << "\tnone\n";
                    else
                        std::cout << "\n";
#define VALUE(v,x) "  " << v << ":" << std::setw(3) << x
#define AMMO(t,u) (t(u).getTorpType() ? t(u).getTorpCount() : t(u).getFighterCount())
                    std::cout << std::setfill(' ')
                              << VALUE("S", result(0).getShield())
                              << VALUE("D", result(0).getDamage())
                              << VALUE("C", result(0).getCrew())
                              << VALUE("A", AMMO(result, 0))
                              << "   |   "
                              << VALUE("S", result(1).getShield())
                              << VALUE("D", result(1).getDamage())
                              << VALUE("C", result(1).getCrew())
                              << VALUE("A", AMMO(result, 1))
                              << "\n";
#undef VALUE
                }
            }
        }
    }
    catch (FileProblemException& e) {
        std::cout << "Exception: " << e.getFileName() << ": " << e.what() << "\n";
    }
    catch (std::exception& e) {
        std::cout << "Exception: " << e.what() << "\n";
    }
}
 
