/**
  *  \file ui/listbox.cc
  *  \brief Listbox Widgets
  *
  *  Widgets displaying (vertical) lists, and related things.
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <cctype>
#include "ui/listbox.h"
#include "ui/uiglobal.h"
#include "ui/commands.h"
#include "gfx/palette.h"
#include "cpluslib/assert.h"
#include "ui/layout.h"
#include "util/nls.h"
#include "gfx/gfxcompl.h"
#include "ui/util.h"
#include "gfx/clip.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/button.h"
#include "client/widgets/helpwidget.h"
#include "ui/widgets/quit.h"

/****************************** UIScrollable *****************************/

/** \class UIScrollable
    \brief Scrollable thing

    This class provides basic operations on a scrollable thing.
    A scrollable thing has the following properties:

    - a current position (cursor)
    - a size
    - a window on the whole data set (top position, page size)

    \invariant 0 <= pos < size */

UIScrollable::UIScrollable(int id)
    : UIWidget(id), pos(0), top(0), size(0), page(1)
{ }

UIScrollable::~UIScrollable()
{ }

/** \function UIScrollable::scrollTo(int pos, int corr)
    Set current position. Derived classes must override this to
    implement the actual scrolling policy.
    \param pos   desired new selected element
    \param corr  (-1 or +1) if object pos can't be selected,
                 correct by going into specified direction */

/** Adjust current position to account for changes in underlying
    data. This makes sure \c pos is actually visible. */
void
UIScrollable::adjustPosition()
{
    if (pos >= size)
        pos = size - 1;
    if (pos < 0)
        pos = 0;
    if (pos < top)
        top = pos;
    if (top == pos && top > 0 && !isAccessible(top-1))
        --top;
    if (pos >= top + page)
        top = pos - page + 1;

    if (size >= page) {
        if (top + page >= size)
            top = size - page;
    }
}

/** Check whether element is accessible. Derived classes can override
    this to implement inaccessible list items.
    \return true if object can be selected by user
    \pre 0 <= arg < getSize() */
bool
UIScrollable::isAccessible(int)
{
    return true;
}

/******************************* UIListbox *******************************/

/** \class UIListbox
    \brief Listbox Widget

    This is a standard list box widget. It provides the basic user interface
    for scrolling around. Derived classes should override drawPart() to
    actually draw the list box.

    The default height of a list cell is the height of FONT_NORMAL; you
    can change that (from the constructor) using setItemHeight(). */

/** Constructor.
    \param id     widget Id
    \param lines  the preferred line count. May be -1 to use the
                  maximum (see setDefaultHeight()).
    \param width  the preferred line width in pixels
    \param items  initial item count */
UIListbox::UIListbox(int id, int lines, int width, int items)
    : UIScrollable(id), lines(lines), width(width),
      item_height(font_heights[FONT_NORMAL]), header_height(0),
      mdown(false),
      have_menu_pos(false)
{
    setSize(items);
    if (lines < 0)
        setDefaultHeight();
}

/** Compute the default (=maximum) height. This computes a nice value
    for /lines/ so that the widget fits on one screen. This does not
    immediately affect display; it just sets the parameters for the
    next layout run. */
void
UIListbox::setDefaultHeight()
{
    int max_lines = (ui_root->getExtent().h - getHeaderHeight()) / getItemHeight() - 10;

    lines = getSize();
    if (lines < 3)
        lines = 3;
    else if (lines > max_lines)
        lines = max_lines;
}

/** Scroll to position. This accounts for disabled items.
    \param pos    desired position
    \param corr   correction, +1 or -1. When the desired position is
                  not accessible, pos is adjusted in that direction. */
void
UIListbox::scrollTo(int pos, int corr)
{
    ASSERT(corr);
    if (pos >= getSize())
        pos = getSize() - 1;
    if (pos < 0)
        pos = 0;
    if (getSize() == 0)
        return;

    /* figure out new position */
    int n = 0;
    while (!isAccessible(pos)) {
        if (pos <= 0)
            corr = 1, ++n;
        else if (pos >= getSize()-1)
            corr = -1, ++n;
        if (n > 2)
            return;
        pos += corr;
    }

    ASSERT(pos >= 0 && pos < getSize() && isAccessible(pos));

    int ot = getTop(), op = getCurrent();
    setCurrent(pos);
    adjustPosition();
    if (ot == getTop() && op == getCurrent())
        return;

    if (!getParent() || !hasState(st_Exposed))
        return;

    sig_moved.raise(getId());
    if (ot == getTop()) {
        int a, b;
        ObjectHolder h;
        GfxCanvas& can = getParent()->getCanvasFor(*this, h);
        can.lock();
        a = op - ot;
        b = getCurrent() - ot;
        if (a > b)
            std::swap(a, b);
        ++b;
        drawPart(can, a, b);
        can.unlock();
        can.update(GfxRect(extent.x, extent.y + header_height + a*item_height,
                           extent.w, (b-a)*item_height));
    } else {
        drawWidget();
    }
}

void
UIListbox::drawContent(GfxCanvas& can)
{
    int partial = (extent.h - header_height) % item_height;
    if (partial != 0) {
        /* Partial line at end */
        GfxClipRectFilter filter(getExtent(), can);
        drawPart(filter, 0, getPageSize()+1);
    } else {
        /* Fits evenly */
        drawPart(can, 0, getPageSize());
    }
    can.update(getExtent());
}

void
UIListbox::onStateChange(int state, bool enable)
{
    UIScrollable::onStateChange(state, enable);
    if (state & st_Focused)
        updateCurrent();
}

bool
UIListbox::handleEvent(const UIEvent& e, bool second_pass)
{
    int fuzz;
    switch (e.type) {
     case ev_Mouse:
        if (hasState(st_Disabled))
            break;

        /* recognize clicks which are doubtlessly inside, and clicks which
           are inside with fuzz factor, for scrolling */
        fuzz = (second_pass && !hasState(lb_Menu)) ? 8 : 0;
        if ((e.mouse.b || hasState(lb_Menu))
            && e.mouse.x >= extent.x && e.mouse.x < extent.x + extent.w
            && e.mouse.y >= extent.y + header_height - fuzz && e.mouse.y < extent.y + extent.h + fuzz)
        {
            setSelectedWidget();
            setFocusedWidget();

            if (e.mouse.y < extent.y + header_height) {
                /* scroll up */
                scrollTo(getTop()-1, -1);
            } else if (e.mouse.y >= extent.y + extent.h) {
                /* scroll down */
                scrollTo(getTop()+getPageSize(), +1);
            } else {
                int y = getTop() + (e.mouse.y - (extent.y + header_height)) / getItemHeight();
                if (isAccessible(y))
                    scrollTo(y, 1);
            }

            if (hasState(lb_Menu)) {
                /* menu: cursor follows mouse, single click posts event */
                if (!e.mouse.b) {
                    if (mdown)
                        sig_dblclick.raise(getId());
                    mdown = false;
                } else {
                    mdown = true;
                }
            } else if (e.mouse.b & mb_Right) {
                have_menu_pos = true;
                menu_pos.x = e.mouse.x;
                menu_pos.y = e.mouse.y;
                sig_menu.raise(getId());
            } else if (e.mouse.b & mb_Double) {
                sig_dblclick.raise(getId());
                mdown = false;
            } else {
                mdown = true;
            }
            return true;
        } else {
            /* Mouse released in a non-menu: generate onItemClick,
               but continue normal processing. */
            if (mdown && isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
                setSelectedWidget();
                int y = getTop() + (e.mouse.y - (extent.y + header_height)) / getItemHeight();
                if (isAccessible(y) && y < getSize()) {
                    scrollTo(y, 1);
                    onItemClick(y, e.mouse.x - getExtent().x);
                }
            }
            mdown = false;
        }
        break;
     case ev_Key:
        if (hasState(st_Disabled) || !hasState(st_Focused))
            break;
        switch (e.key.code) {
         case SDLK_UP:
         case vk_WheelUp:
            setSelectedWidget();
            scrollTo(getCurrent()-1, -1);
            return true;
         case SDLK_DOWN:
         case vk_WheelDown:
            setSelectedWidget();
            scrollTo(getCurrent()+1, +1);
            return true;
         case SDLK_HOME:
            setSelectedWidget();
            scrollTo(0, 1);
            return true;
         case SDLK_END:
            setSelectedWidget();
            scrollTo(getSize()-1, -1);
            return true;
         case SDLK_PAGEUP:
            if (hasState(lb_NoPageKeys))
                break;
            /* FALLTHROUGH */
         case ss_Shift + SDLK_PAGEUP:
            setSelectedWidget();
            scrollTo(getCurrent() - getPageSize(), -1);
            return true;
         case SDLK_PAGEDOWN:
            if (hasState(lb_NoPageKeys))
                break;
            /* FALLTHROUGH */
         case ss_Shift + SDLK_PAGEDOWN:
            setSelectedWidget();
            scrollTo(getCurrent() + getPageSize(), -1);
            return true;
         case '#':
         case '\\':
         case SDLK_MENU:        // why not
            if (hasState(lb_KeyboardMenu)) {
                setSelectedWidget();
                sig_menu.raise(getId());
                return true;
            }
            break;
            // c-pgup, c-pgdn, c-home, c-end
        }
        break;
     default:;
    }
    return UIScrollable::handleEvent(e, second_pass);
}

bool
UIListbox::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_INT_QueryReferencePoint) {
        // Menu position query
        ASSERT(cmd.infop);
        TReferencePointArg& ra = *static_cast<TReferencePointArg*>(cmd.infop);
        if (have_menu_pos) {
            // We have a menu position
            ra.x = menu_pos.x;
            ra.y = menu_pos.y;
        } else {
            // Default position
            ra.x = getExtent().x + 10;
            ra.y = getExtent().y + getExtent().h / 2;
        }
        return true;
    } else {
        // Unsupported command
        return UIScrollable::handleCommand(cmd, sender);
    }
}

void
UIListbox::getLayoutInfo(LayoutInfo& info)
{
    int minLines, prefLines;
    if (lines == 0) {
        minLines = prefLines = 1;
    } else if (lines < 3) {
        minLines = prefLines = lines;
    } else {
        minLines = 3;
        prefLines = lines;
    }

    info.flags     = li_Flexible;
    info.min_size  = GfxPoint(width, item_height * minLines + header_height);
    info.pref_size = GfxPoint(width, item_height * prefLines + header_height);
}

void
UIListbox::onItemClick(int /*item*/, int /*x*/)
{ }

void
UIListbox::onResize(GfxRect oldsize, bool by_layout)
{
    UIScrollable::onResize(oldsize, by_layout);

    // Cancel old menu position
    have_menu_pos = false;

    GfxRect newpos = getExtent();
    setPageSize((newpos.h - header_height) / item_height);
    if (getPageSize() >= getSize()) {
        setTop(0);
    } else {
        /* try to put Top so that Pos is in center */
        int n = getCurrent() - getPageSize() / 2;
        if (n < 0)
            setTop(0);
        else
            setTop(n);
    }
    adjustPosition();
    sig_moved.raise(getId());
}

/** Redraw the current entry of the list box. */
void
UIListbox::updateCurrent()
{
    if (!getParent() || !hasState(st_Exposed))
        return;

    int a = getCurrent() - getTop(), b = a+1;
    ObjectHolder h;
    GfxCanvas& can = getParent()->getCanvasFor(*this, h);
    can.lock();
    drawPart(can, a, b);
    can.unlock();
    can.update(GfxRect(extent.x, extent.y + a*item_height + header_height,
                       extent.w, (b-a)*item_height));
}

/** Prepare for drawing an entry.
    \param ctx      [in/out] Graphics context
    \param style    [in] Style of this list box
    \param line     [in] Logical line to draw [0, getPageSize())
    \param area_out [out] If non-null, area of the item is returned here */
void
UIListbox::prepareEntry(GfxContext& ctx, Style style, int line, GfxRect* area_out)
{
    const int real_y = item_height*line + extent.y + header_height;
    GfxRect area(extent.x, real_y, extent.w, item_height);
    if (style == HighContrastStyle) {
        /* High contrast: */
        ctx.useColorScheme(getSkin());
        if (hasState(st_Disabled) || !isAccessible(line + getTop())) {
            drawBackground(ctx, area);
            ctx.setColor(UIColor::tc_Faded);
        } else if (line + getTop() == getCurrent()) {
            if (getFocusState() == NoFocus) {
                drawBackground(ctx, area);
                ctx.setColor(UIColor::tc_Static);
                drawRectangle(ctx, area);
            } else {
                drawSolidBar(ctx, area, UIColor::tc_Static);
                ctx.setColor(UIColor::tc_InvStatic);
            }
        } else {
            drawBackground(ctx, area);
            ctx.setColor(UIColor::tc_Static);
        }
    } else {
        /* Colored: */
        /* FIXME: can this be made more elegant? */
        if (line + getTop() == getCurrent() && style != CursorlessColorStyle) {
            if (getFocusState() == NoFocus) {
                static UISkin sk(BRIGHT_GRAY_COLOR_SET);
                ctx.useColorScheme(sk);
            } else {
                static UISkin sk(BLACK_COLOR_SET);
                ctx.useColorScheme(sk);
            }
        } else {
            static UISkin sk(GRAY_COLOR_SET);
            ctx.useColorScheme(sk);
        }
        if (isAccessible(line + getTop()) && (!hasState(st_Disabled) || style == CursorlessColorStyle)) {
            ctx.setColor(UIColor::tc_Static);
        } else {
            ctx.setColor(UIColor::tc_Faded);
        }
        drawBackground(ctx, area);
    }
    if (area_out)
        *area_out = area;
}

/*************************** UIStandardListbox ***************************/

/** \class UIStandardListbox
    \brief Standard List Box Widget

    This class provides the standard list box widget used for most cases.
    It displays a UIStringList, and provides handy functions to access
    the UIStringList's keys.

    In a UIStandardListbox, users can type the first letter of an entry
    for "quick search". */

UIStandardListbox::UIStandardListbox(int id, int lines, int width, const UIStringList& data)
    : UIListbox(id, lines, width, data.getCount()), data(data), tab_width(0)
{ }

UIStandardListbox::~UIStandardListbox()
{ }

void
UIStandardListbox::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_NORMAL]);
    UIStringList::iterator it = data.getItem(from + getTop());
    for (int i = from; i < to; ++i) {
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        if (it != data.end()) {
            string_t::size_type n;
            if (tab_width > 0 && tab_width < area.w-5 && (n = it->second.find('\t')) != string_t::npos) {
                outTextF(ctx, area.x+5, area.y, tab_width, it->second.substr(0, n));
                outTextF(ctx, area.x+5+tab_width, area.y, area.w-5-tab_width, it->second.substr(n+1));
            } else {
                outTextF(ctx, area.x+5, area.y, area.w-5, it->second);
            }
            ++it;
        }
    }
}

/** Scroll to the item which has the specified key. Does nothing if no
    such element exists or it is not accessible. */
void
UIStandardListbox::scrollToKey(int key)
{
    int n = data.getItemIndexByKey(key);
    if (n >= 0 && isAccessible(key))
        scrollTo(n, 1);
}

bool
UIStandardListbox::handleEvent(const UIEvent& e, bool second_pass)
{
    if (getFocusState() == PrimaryFocus
        && e.type == ev_Key
        && (e.key.code >= ' ' && e.key.code < 127)
        && data.getCount() > 0)
    {
        /* FIXME: this assumes ASCII, like large parts of this
           program. Should we support umlauts, etc? */
        using namespace std;
        char c = toupper(e.key.code);

        int pos = getCurrent();
        bool wrap = false;
        while (1) {
            ++pos;
            if (pos >= data.getCount()) {
                if (wrap)
                    break;
                pos = 0;
                wrap = true;
            }

            UIStringList::iterator it = data.getItem(pos);
            if (isAccessible(pos)
                && it->second.length()
                && toupper(static_cast<unsigned char>(it->second[0])) == c)
            {
                /* found one */
                scrollTo(pos, 1);
                return true;
            }
        }
    }

    return UIListbox::handleEvent(e, second_pass);
}

/**************************** Standard Dialogs ***************************/

/** Standard list box dialog. Displays the widget in a standard modal
    dialog window.
    \return true if user chose "OK", false otherwise. */
bool
UIListbox::doStandardListbox(string_t title, string_t help)
{
    ObjectHolder h;

    UIWindow& w = h.add(new UIWindow(title, UIWindow::BLUE_WINDOW, COLOR_GRAY));
    w.setLayout(UIVBoxLayout::instance5);

    UIFrameGroup& list_group = h.add(new UIFrameGroup(UIHBoxLayout::instance0));
    list_group.add(*this);
    w.add(list_group);
    // FIXME: the following assumes we get as many lines from layout as we request.
    if (getSize() > lines)
        list_group.add(h.add(new UIScrollbar(*this)));

    UIGroup& button_group = h.add(new UIGroup(UIHBoxLayout::instance5));
    button_group.add(h.add(new UIButton(cm_Enter,   _("OK"), SDLK_RETURN)).addStop());
    button_group.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    button_group.add(h.add(new UISpacer()));
    if (!help.empty()) {
        button_group.add(h.add(new UIButton(_("H"), 'H')));
        w.add(h.add(new WHelpWidget(help)));
    }
    w.add(button_group);
    w.add(h.add(new UIQuit(cm_Escape)));

    w.pack();
    ui_root->centerWidget(w);
    return ui_root->runModal(w) == cm_Enter;
}

namespace {
    class MenuFrame : public UIFrameGroup {
        int target_height;
        long next_tick;
        enum { TIMER_DIVI = 0 };
     public:
        MenuFrame(int target_height);
        ~MenuFrame();
        void drawContent(GfxCanvas& can);
        bool handleEvent(const UIEvent& e, bool second_pass);
        void stopValid(int) { stop(1); }
    };
}

MenuFrame::MenuFrame(int target_height)
    : UIFrameGroup(UIHBoxLayout::instance0, 2), target_height(target_height),
      next_tick(getTimer() + TIMER_DIVI)
{
    setState(st_Modal, true);
}

MenuFrame::~MenuFrame()
{ }

void
MenuFrame::drawContent(GfxCanvas& can)
{
    // FIXME: I don't like this. Should probably not be based on UIFrameGroup.
    GfxRect out = getExtent(), in = getExtent();
    adjustSize(in, true);
    in.grow(1, 1);

    drawFrameUp(can, out);
    drawFrameDown(can, in);
    can.update(getExtent());
}

bool
MenuFrame::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Tick && getTimer() > next_tick) {
        next_tick = getTimer() + TIMER_DIVI;
        GfxRect ex = getExtent();
        if (ex.h < target_height) {
            ex.h += font_heights[FONT_NORMAL];
            if (ex.h > target_height)
                ex.h = target_height;
            ex.moveInto(ui_root->getExtent());
            setExtent(ex, false);
        }
    } else if (e.type == ev_Key) {
        if (e.key.code == SDLK_ESCAPE) {
            stop(0);
            return true;
        } else if (e.key.code == SDLK_RETURN) {
            stop(1);
            return true;
        }
    } else if (e.type == ev_Mouse && e.mouse.b && !isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
        ungetEvent(e);
        stop(0);
        return true;
    }
    return UIFrameGroup::handleEvent(e, second_pass);
}

/** Displays the list as a standard menu. This will "fade in" the
    menu, and let the user choose from it.
    \param x,y left-top position of menu
    \return true if user chose an item, false if not. */
bool
UIListbox::doStandardMenu(int x, int y)
{
    GfxRect menu_extent(x, y, width+4, font_heights[FONT_NORMAL] + 4);

    int maxLines = (ui_screen->getSize().y - y - 4) / font_heights[FONT_NORMAL];
    if (maxLines > 20) {
        maxLines = 20;
    }
    if (maxLines < 8) {
        maxLines = 8;
    }

    int menuHeight = std::min(maxLines, getSize());
    MenuFrame mf(font_heights[FONT_NORMAL]*menuHeight + 4 /* FIXME */);
    SignalConnection keks(sig_dblclick.add(&mf, &MenuFrame::stopValid));
    setState(lb_Menu, true);

    // The Menu
    mf.add(*this);

    // The Scrollbar
    ObjectHolder h;
    if (menuHeight < getSize()) {
        mf.add(h.add(new UIScrollbar(*this)));
        menu_extent.w += 20;
    }

    // Size
    menu_extent.moveInto(ui_root->getExtent());
    mf.setExtent(menu_extent, false);
    mf.add(h.add(new UIQuit(cm_Escape)));
    ui_root->add(mf);
    int i = ui_root->run();
    ui_root->close(mf);
    keks.disconnect();          // not actually needed
    return i != 0;
}

/** Displays the list as a standard menu. Like
    doStandardMenu(int,int), but this will query the display position
    using the specified reference event (in plain-text, it will send a
    cm_INT_QueryReferencePoint aka "who sent this" message and use the
    return value). */
bool
UIListbox::doStandardMenu(UIBaseWidget& ref_widget)
{
    TReferencePointArg rpa;
    UICommand cmd(cm_INT_QueryReferencePoint);
    cmd.infop = &rpa;
    rpa.x = ref_widget.getExtent().x;
    rpa.y = ref_widget.getExtent().getBottomY();
    ref_widget.handleCommand(cmd, *this);
    ref_widget.handleCommand(UICommand(cm_INT_Push), *this);
    bool rv = doStandardMenu(rpa.x, rpa.y);
    ref_widget.handleCommand(UICommand(cm_INT_Unpush), *this);
    return rv;
}

/****************************** UIScrollbar ******************************/

UIScrollbar::UIScrollbar(UIScrollable& scrl)
    : UIWidget(), scrl(scrl), conn(scrl.sig_moved.add(this, &UIScrollbar::scrolled))
{
    sync();
}

UIScrollbar::~UIScrollbar()
{ }

void
UIScrollbar::scrolled(int)
{
    drawWidget(false);
}

void
UIScrollbar::sync()
{
    if (scrl.getPageSize() >= scrl.getSize()) {
        /* no scrollbar needed */
        bf[0] = bf[1] = st_Disabled;
    } else {
        /* scrollbar needed */
        bf[0] = bf[1] = 0;
    }
    drawWidget(false);
}

void
UIScrollbar::setBF(int id, int flags)
{
    if (flags != bf[id]) {
        bf[id] = flags;
        drawWidget(false);
    }
}

void
UIScrollbar::drawContent(GfxCanvas& can)
{
    GfxRect interior = extent;

    /* buttons */
    if (interior.h >= 20) {
        UIButton::drawButton(can, GfxRect(extent.x, extent.y,
                                          extent.w, 20), "\xE2\x86\x91", bf[0]);
        UIButton::drawButton(can, GfxRect(extent.x, extent.y + extent.h - 20,
                                          extent.w, 20), "\xE2\x86\x93", bf[1]);
    }

    /* scrollbox */
    if (interior.h >= 42) {
        interior.grow(0, -20);
        drawFrameUp(can, interior);
        interior.grow(-1, -1);

        if (scrl.getSize() > scrl.getPageSize()) {
            /* draw scroll box */
            int top = scrl.getTop();
            if (top + scrl.getPageSize() > scrl.getSize())
                top = scrl.getSize() - scrl.getPageSize();
            int divi = scrl.getSize();

            int yy1 = interior.h * int32(top) / divi;
            int yy2 = interior.h * int32(top + scrl.getPageSize()) / divi;

            if (yy1 > 0)
                drawSolidBar(can, GfxRect(interior.x, interior.y, interior.w, yy1),
                             standard_colors[COLOR_SHIELD + 3]);
            drawFrameUp(can, GfxRect(interior.x, interior.y + yy1, interior.w, yy2 - yy1));
            if (yy1 < yy2-2)
                drawSolidBar(can, GfxRect(interior.x+1, interior.y + yy1+1,
                                          interior.w-2, yy2-yy1-2),
                             standard_colors[COLOR_SHIELD + 1]);
            if (yy2 < interior.h)
                drawSolidBar(can, GfxRect(interior.x, interior.y+yy2,
                                          interior.w, interior.h-yy2),
                             standard_colors[COLOR_SHIELD + 3]);
        } else {
            drawSolidBar(can, interior, standard_colors[COLOR_SHIELD + 5]);
        }
    }

    can.update(extent);
}

void
UIScrollbar::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_FlexibleV;
    info.min_size = GfxPoint(20, 60);
    info.pref_size = GfxPoint(20, 80);
}

bool
UIScrollbar::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Mouse) {
        int old_bf1 = bf[0], old_bf2 = bf[1];
        bf[0] &= ~(st_Selected | UIButton::bf_Pressed);
        bf[1] &= ~(st_Selected | UIButton::bf_Pressed);
        if (isInside(GfxPoint(e.mouse.x, e.mouse.y))) {
            int where = -1;
            if (e.mouse.y < extent.y + 20)
                where = 0;
            else if (e.mouse.y >= extent.y + extent.h - 20)
                where = 1;

            if (where >= 0) {
                setSelectedWidget();
                bf[where] |= st_Selected;
                if (e.mouse.b)
                    bf[where] |= UIButton::bf_Pressed;
            }
        }
        if (bf[0] != old_bf1 || bf[1] != old_bf2) {
            drawWidget(false);
            return true;
        }
        /* FALLTHROUGH */
    } else if (e.type == ev_Tick && hasState(st_Selected)) {
        if (bf[0] & UIButton::bf_Pressed) {
            scrl.setFocusedWidget();
            scrl.scrollTo(scrl.getTop() - 1, -1);
        } else if (bf[1] & UIButton::bf_Pressed) {
            scrl.setFocusedWidget();
            scrl.scrollTo(scrl.getTop() + scrl.getPageSize(), +1);
        }
        /* FALLTHROUGH */
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
UIScrollbar::onStateChange(int state, bool enable)
{
    if ((state & st_Selected) && !enable) {
        setBF(0, bf[0] & ~(st_Selected | UIButton::bf_Pressed));
        setBF(1, bf[1] & ~(st_Selected | UIButton::bf_Pressed));
    }
    UIWidget::onStateChange(state, enable);
}
