/**
  *  \file misc/random.cc
  *  \brief Random Numbers
  *
  *  Instead of relying on the vague semantics of std::rand(), we use
  *  our own, simple random number generator. This one is expected to
  *  have a period of 4G (in other words, enough for us).
  *
  *  This generator can be instantiated at will, and there also is a
  *  global instance. Note that this is called `myRandom' because
  *  there is already a `random()' function in BSD. Use your own
  *  instances if you need a pseudo-random stream for otherwise
  *  deterministic behaviour (e.g. TRN encryption), use the global
  *  instance for general "nondeterminism" (e.g. visual effects).
  *
  *  (c) 2001-2019 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <ctime>
#include "util/random.h"

/************************* RandomNumberGenerator *************************/

RandomNumberGenerator::RandomNumberGenerator(uint32 seed)
    : seed(seed)
{ }

/** Advance seed. Simple linear congruential sequence. */
inline void
RandomNumberGenerator::next()
{
    seed = 134775813UL * seed + 1;
}

/** Get random number in range [0, 2^16). */
uint16
RandomNumberGenerator::operator()()
{
    next();
    return seed >> 16;
}

/** Get random number in range [0, max). */
uint16
RandomNumberGenerator::operator()(uint16 max)
{
    next();
    /* seed * max >> 32        
        = (loword(seed) * max + hiword(seed) * max << 16) >> 32
          hence loword(result) = loword(seed)*max which is discarded anyways
        = (loword(seed) * max >> 16 + hiword(seed) * max) >> 16 */
    return (((seed & 0xFFFFU) * max >> 16) + (seed >> 16) * max) >> 16;
}

/** Global random number generator instance.
    (This used to be a set of overloaded functions) */
RandomNumberGenerator myRandom(std::time(0));
